#include <fel.h>
#include <spinand.h>


static struct xfel_ctx_t ctx;
static char Name[128];
static uint32_t Capacity, Read; 
static char *flashbf, *filebf;


static int terminal_error(void){
  if(ctx.hdl)
    libusb_close(ctx.hdl);
	libusb_exit(NULL);
  if(flashbf)
    free(flashbf);
  if(filebf)
    free(filebf);
  exit(-1);  
}
static uint32_t file_save(const char * filename, void * buf, uint32_t len)
{
	FILE * out = fopen(filename, "wb");
	int r;
	if(!out)
	{
    return 0;
	}
	r = fwrite(buf, len, 1, out);
	fclose(out);
	return r;
}

static void * file_load(const char * filename, uint32_t * len)
{
  uint32_t size, n;
  FILE * in;
  char * buf;
  in = fopen(filename, "rb");
    
  fseek(in, 0, SEEK_END);       // seek to end of file
  size = ftell(in);             // get current file pointer
  fseek(in, 0, SEEK_SET);       // seek back to beginning of file  
	buf = malloc(size);        // allocate size
  if(!buf){
    printf("Unable to allocate file buffer!\r\n");
    return NULL;
  }
  n = fread(buf, 1, size, in);  // Read whole file
  
  if(n < size){                 // Ensure it was completely read  
      free(buf);
      return NULL;
  }
  
	if(len)
		*len = n;
  
	if(in != stdin)
		fclose(in);
	return buf;
}


static void usage(void)
{
	printf("\r\nHantek DSO2D1x flash utility v0.2\r\n");
  printf("Based on xfel v1.2.6 - https://github.com/xboot/xfel\r\n");
	printf("Usage:\r\n");
	printf("    dsoflash -h or --help                         - Show this screen\r\n");
	printf("    dsoflash ver                                  - Get chip version\r\n");
	printf("    dsoflash detect                               - Detect flash\r\n");
	printf("    dsoflash status                               - Get flash status regs\r\n");
	printf("    dsoflash read <file>                          - Dump flash to file\r\n");
	printf("    dsoflash write <file>                         - Restore flash from file\r\n");
	printf("    dsoflash erase                                - Erase flash\r\n\n");
	printf("Warning: Commands will be executed inmediately, without confirmation!\r\n");
}

int main(int argc, char * argv[])
{
	if(argc < 2)
	{
		usage();
		return 0;
	}
  
	argc --;
	argv ++;
	for(int i = 1; i < argc; i++)
	{
		if(!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help"))
		{
			usage();
			return 0;
		}
	}

	libusb_init(NULL);  
	ctx.hdl = libusb_open_device_with_vid_pid(NULL, 0x1f3a, 0xefe8);
	if(!fel_init(&ctx))
	{
		printf("ERROR: Can't found any FEL device\r\n");
		if(ctx.hdl)
			libusb_close(ctx.hdl);
		libusb_exit(NULL);
		return -1;
	}
   
	if(!strcmp(argv[0], "ver"))
	{
		printf("%.8s ID=0x%08x(%s) dflag=0x%02x dlength=0x%02x scratchpad=0x%08x\r\n",
			ctx.version.magic, ctx.version.id, ctx.chip->name, ctx.version.dflag,
			ctx.version.dlength, ctx.version.scratchpad);
	}
    
	else if(!strcmp(argv[0], "detect") && (argc == 1))
	{
    if(spinand_detect(&ctx, Name, &Capacity))
      printf("Found spi nand flash '%s' with %u bytes\r\n", Name, Capacity);
    else
      printf("Unknown flash memory!\r\n");
  }
  
  else if(!strcmp(argv[0], "status") && (argc == 1))
    dso2d_dump_regs(&ctx);

  else if(!strcmp(argv[0], "erase") && (argc == 1))
    dso2d_erase(&ctx);

  else if(!strcmp(argv[0], "read") && (argc == 2))
  {
    if(!spinand_detect(&ctx, Name, &Capacity))
    {
      printf("Unknown flash memory!\r\n");
      terminal_error();
    }
    else
      printf("Flash found: '%s'  Size: %u bytes\r\n", Name, Capacity);
    
    flashbf = malloc(Capacity);
    if(!flashbf)
    {
      printf("Unable to allocate flash buffer!\r\n");
      terminal_error();        
    }
    dso2d_dump(&ctx, flashbf);
    if(!file_save(argv[1], flashbf, Capacity))
    {
      printf("Unable to write to file \"%s\"!\r\n", argv[1]);
      terminal_error();              
    }
    else
    {
      printf("Flash read sucessfully into file \"%s\"\r\n", argv[1]);
      free(flashbf);
    }
  }
  
  else if(!strcmp(argv[0], "write") && (argc == 2))
  {
    if(!spinand_detect(&ctx, Name, &Capacity))
    {
      printf("Unknown flash memory!\r\n");
      terminal_error();
    }
    else
      printf("Flash found: '%s'  Size: %u bytes\r\n", Name, Capacity);
    
    filebf = file_load(argv[1], &Read);
    if(!filebf)
    {
      printf("Unable to read from file \"%s\"!\r\n", argv[1]);
      terminal_error();      
    }
    if (Read!=Capacity)
    {          
      printf("File doesn't match the flash size\r\n");
      printf(" Flash: %uBytes,   File: %uBytes\r\n", Read, Capacity);
      terminal_error();
    }
    else
    {
      dso2d_restore(&ctx, filebf);
      printf("Flash written sucessfully from file \"%s\"\r\n", argv[1]);
      free(filebf);
    }
  }
  else
    usage();
  
	if(ctx.hdl)
		libusb_close(ctx.hdl);
	libusb_exit(NULL);

	return 0;
}
